/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_LCO_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void LengthCorrectOp::initialize(void)
{
	catalog<I32>("Iterations")=1;

	catalog<bool>("Root Lock")=false;
	catalog<String>("Root Lock","hint")=
			"Leave first vertex of each primitive untouched."
			"  This can lock down rooted curves, like hair.";

	catalog<Real>("Correction Bias")=0.5;
	catalog<Real>("Correction Bias","max")=1.0;
	catalog<String>("Correction Bias","hint")=
			"Influence how much each point on an edge gets changed."
			"  Try lower numbers for rooted curves, like 0.3,"
			" but generally leave at 0.5 for polygons.";

	catalog<Real>("Suspension")=0.0;
	catalog<Real>("Suspension","max")=1.0;
	catalog<String>("Suspension","hint")=
			"Restrain changes to slide along original input.";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Reference Surface");
}

void LengthCorrectOp::handle(Record& a_rSignal)
{
#if FE_LCO_DEBUG
	feLog("LengthCorrectOp::handle\n");
#endif

	const I32 iterations=catalog<I32>("Iterations");
	const BWORD rootLock=catalog<bool>("Root Lock");
	const Real bias=catalog<Real>("Correction Bias");
	const Real suspension=catalog<Real>("Suspension");

	//* output surface is copy of input surface
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	//* access reference surface
	sp<SurfaceAccessibleI> spReferenceAccessible;
	if(!access(spReferenceAccessible,"Reference Surface")) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spRefVertices;
	if(!access(spRefVertices,spReferenceAccessible,
			e_primitive,e_vertices)) return;

	//* TODO need special case for curves
	if(suspension>0.0)
	{
		//* access input surface
		sp<SurfaceAccessibleI> spInputAccessible;
		if(!access(spInputAccessible,"Input Surface")) return;

		if(!access(m_spDriver,spInputAccessible)) return;

		m_spDriver->setRefinement(0);
		m_spDriver->prepareForSearch();
	}
	else
	{
		m_spDriver=NULL;
	}

	const Real antibias=1.0-bias;

	//* tweak surface points towards reference length
	const U32 count=spOutputVertices->count();
	const U32 countRef=spRefVertices->count();
	if(count!=countRef)
	{
		feLog("LengthCorrectOp::handle mismatched count %d vs %d ref\n",
				count,countRef);
		catalog<String>("error")="mismatched reference surface";
		return;
	}

#if FE_LCO_DEBUG
	feLog("LengthCorrectOp::handle primitives %d\n", count);
#endif

	for(U32 iteration=0;iteration<iterations;iteration++)
	{
		for(U32 index=0;index<count;index++)
		{
			const U32 subCount=spOutputVertices->subCount(index);
			const U32 subCountRef=spRefVertices->subCount(index);
			if(subCount!=subCountRef)
			{
				feLog("LengthCorrectOp::handle"
						" primitive %d mismatched count %d vs %d ref\n",
						index,subCount,subCountRef);
				continue;
			}

#if FE_LCO_DEBUG
			feLog("LengthCorrectOp::handle iteration %d/%d"
					" primitive %d/%d subCount %d\n",
					iteration,iterations,index,count, subCount);
#endif

			//* NOTE skip root point with rootLock
			for(U32 subIndex=rootLock;subIndex<subCount;subIndex++)
			{
				const U32 lastSubIndex=subIndex? subIndex-1: subCount-1;
				SpatialVector pointA=
						spOutputVertices->spatialVector(index,lastSubIndex);
				SpatialVector pointB=
						spOutputVertices->spatialVector(index,subIndex);

#if FE_LCO_DEBUG
				feLog("LengthCorrectOp::handle   %d/%d pointA %s pointB %s\n",
						subIndex,subCount,c_print(pointA),c_print(pointB));
#endif

				const SpatialVector delta=pointB-pointA;
				const Real length=magnitude(delta);
				if(length>0.0)
				{
					const SpatialVector pointRefA=
							spRefVertices->spatialVector(index,lastSubIndex);
					const SpatialVector pointRefB=
							spRefVertices->spatialVector(index,subIndex);

					const Real lengthRef=magnitude(pointRefB-pointRefA);
					const SpatialVector adjust=(lengthRef/length-1.0)*delta;

#if FE_LCO_DEBUG
					feLog("LengthCorrectOp::handle   ref pointA %s pointB %s\n",
							c_print(pointRefA),c_print(pointRefB));
#endif

					if(!rootLock || subIndex>1)
					{
						pointA-=bias*adjust;
						pointB+=antibias*adjust;

						spOutputVertices->set(index,lastSubIndex,pointA);
						spOutputVertices->set(index,subIndex,pointB);
					}
					else
					{
						pointB+=adjust;
						spOutputVertices->set(index,subIndex,pointB);
					}

#if FE_LCO_DEBUG
					feLog("LengthCorrectOp::handle   now pointA %s pointB %s\n",
							c_print(pointA),c_print(pointB));
#endif
				}
			}
		}

		if(m_spDriver.isValid())
		{
			setFullRange(0,count-1);
			OperatorThreaded::handle(a_rSignal);

		}
	}

#if FE_LCO_DEBUG
	feLog("LengthCorrectOp::handle done\n");
#endif
}

void LengthCorrectOp::run(I32 a_id,Range a_range)
{
	const BWORD rootLock=catalog<bool>("Root Lock");
	const Real suspension=catalog<Real>("Suspension");

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;

	for(U32 index=start;index<=end;index++)
	{
		const U32 subCount=spOutputVertices->subCount(index);
		//* NOTE skip root point with rootLock
		for(U32 subIndex=rootLock;subIndex<subCount;subIndex++)
		{
			SpatialVector point=
					spOutputVertices->spatialVector(index,subIndex);

			sp<SurfaceI::ImpactI> spImpact=
					m_spDriver->nearestPoint(point);
			if(spImpact.isValid())
			{
				point+=suspension*(spImpact->intersection()-point);
				spOutputVertices->set(index,subIndex,point);
			}
		}
	}
}
