/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_DODGE_DEBUG	FALSE

#include "grass/grass.pmh"

using namespace fe;

void DodgeOp::initialize(void)
{
	//* parameters
	catalog<Real>("Threshold")=0.1;
	catalog<String>("Threshold","hint")="Gap to maintain around collider.";

	catalog<Real>("Repulsion")=0.1;
	catalog<String>("Repulsion","hint")=
			"Push away nearby curves, even if not colliding.";

	catalog<Real>("Falloff")=1.0;
	catalog<Real>("Falloff","high")=10.0;
	catalog<Real>("Falloff","max")=100.0;
	catalog<String>("Falloff","hint")=
			"Fade repulsion to zero over this distance.";

	catalog<I32>("Refinement")=0;
	catalog<I32>("Refinement","high")=4;
	catalog<I32>("Refinement","max")=8;
	catalog<String>("Refinement","hint")=
			"Maximum level of collider subdivision."
			"  Refinement may be adaptive and non-uniform.";

	catalog<bool>("Debug Output")=false;
	catalog<String>("Debug Output","hint")=
			"Draws lines over the output showing the interactions."
			"  This feedback may only be visible is this node is"
			" \"selected\" and/or \"displayed\".";

	//* input
	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Collider");

	m_spDebugGroup=new DrawMode();
	m_spDebugGroup->setGroup("debug");
}

void DodgeOp::handle(Record& a_rSignal)
{
#if FE_DODGE_DEBUG
	feLog("DodgeOp::handle\n");
#endif

	const I32 refinement=catalog<I32>("Refinement");

	//* output curves is copy of input curves
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceI> spCollider;
	if(!access(spCollider,"Collider")) return;

	spCollider->setRefinement(refinement);
	spCollider->prepareForSearch();

	//* access DrawI interpreter
	const BWORD debug=catalog<bool>("Debug Output");
	if(debug)
	{
		if(!accessGuide(m_spDrawDebug,a_rSignal)) return;
	}
	else
	{
		m_spDrawDebug=NULL;
	}

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->pushDrawMode(m_spDebugGroup);
	}

	setFullRange(0,spOutputVertices->count()-1);
	OperatorThreaded::handle(a_rSignal);

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->popDrawMode();
	}

#if FE_DODGE_DEBUG
	feLog("DodgeOp::handle done\n");
#endif
}

void DodgeOp::run(I32 a_id,Range a_range)
{
	const Color blue(0.0,0.0,1.0);
	const Color yellow(1.0,1.0,0.0);

	const Real threshold=catalog<Real>("Threshold");
	const Real falloff=catalog<Real>("Falloff");
	const Real repulsion=catalog<Real>("Repulsion");

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceI> spCollider;
	if(!access(spCollider,"Collider")) return;

	Real radius=spCollider->radius();
	SpatialVector center=spCollider->center();
	const Real extendedRadius=radius+threshold+falloff;
	const Real extendedRadius2=extendedRadius*extendedRadius;

	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;
	for(U32 index=start;index<=end;index++)
	{
		//* push curve points out of collider in direction of nearest normal

		BWORD hit=FALSE;
		SpatialVector pivotPoint=spOutputVertices->spatialVector(index,0);
		SpatialQuaternion pivotQuat;

		const U32 subCount=spOutputVertices->subCount(index);
		if(!subCount)
		{
			continue;
		}

		const SpatialVector rootPoint=spOutputVertices->spatialVector(index,0);
		const SpatialVector tipPoint=
				spOutputVertices->spatialVector(index,subCount-1);

		const Real approxLength=magnitude(tipPoint-rootPoint);
		const Real scopeRadius=extendedRadius+approxLength;
		const Real scopeRadius2=scopeRadius*scopeRadius;

		if(magnitudeSquared(rootPoint-center)>scopeRadius2)
		{
			continue;
		}

		SpatialVector lastPoint=spOutputVertices->spatialVector(index,0);
		//* NOTE skip root point
		for(U32 subIndex=1;subIndex<subCount;subIndex++)
		{
			BWORD changed=FALSE;
			SpatialVector point=spOutputVertices->spatialVector(index,subIndex);
			const SpatialVector oldPoint=point;

			SpatialVector arm1=point-pivotPoint;
			SpatialVector arm2;
			if(hit)
			{
				rotateVector(pivotQuat,arm1,arm2);
				point=pivotPoint+arm2;
			}

#if FE_DODGE_DEBUG
			feLog("\nDodgeOp::handle %d %d/%d  %s\n",index,
					subIndex,subCount,c_print(point));
			feLog("pivotPoint %s\n",c_print(pivotPoint));
			feLog("pivotQuat %s\n",c_print(pivotQuat));
			feLog("arm1 %s\n",c_print(arm1));
			feLog("arm2 %s\n",c_print(arm2));
#endif

			const Real mag2=magnitudeSquared(point-center);
			if(!hit && mag2>extendedRadius2)
			{
				if(mag2>scopeRadius2)
				{
					subIndex=subCount;
				}
				continue;
			}

			sp<SurfaceI::ImpactI> spImpact=spCollider->nearestPoint(point);
			if(spImpact.isValid())
			{
				const SpatialVector intersection=spImpact->intersection();
				const SpatialVector normal=spImpact->normal();
				const SpatialVector contact=intersection+threshold*normal;
				const SpatialVector reach=
						intersection+(threshold+falloff)*normal;
#if FE_DODGE_DEBUG
				feLog("DodgeOp::handle impact %s\n",
						c_print(intersection));
#endif

				const SpatialVector to_contact=contact-point;
				const SpatialVector to_reach=reach-point;
				const Real len2=magnitudeSquared(to_reach);

				if(len2>1e-6)
				{
					Real len=sqrt(len2);

					const SpatialVector unitReach=to_reach*(1.0/len);
					Real impactDot=dot(unitReach,normal);
					if(impactDot>0.0)
					{
						const SpatialVector unitSegment=
								unitSafe(point-lastPoint);
						const SpatialVector projectionSegment=
								unitSegment*dot(unitSegment,normal);
						const SpatialVector sideRepulsion=
								unitSafe(normal-projectionSegment);
						SpatialVector unitRepulsion=sideRepulsion;

						SpatialVector correction=to_contact;
						Real scalarRepulsion=repulsion;

						BWORD draw=FALSE;
						if(falloff>0.0 && len<falloff)
						{
							set(correction);
							const Real ratio=len/falloff;
							scalarRepulsion*=ratio;
							unitRepulsion=unitSafe(sideRepulsion*ratio+
									normal*(1.0-ratio));
						}
						else
						{
							scalarRepulsion-=impactDot*magnitude(correction);
							draw=TRUE;
						}

						if(scalarRepulsion>0.0)
						{
							correction+=scalarRepulsion*unitRepulsion;
						}

//						correction*=pow(impactDot,0.1);
						point+=correction;
						changed=TRUE;

						if(m_spDrawDebug.isValid() && draw)
						{
							SpatialVector line[2];
							line[0]=point;
							line[1]=point+unitRepulsion;

							Color colors[2];
							colors[0]=blue;
							colors[1]=yellow;

							SAFEGUARD;
							m_spDrawDebug->drawLines(line,NULL,2,
									DrawI::e_discrete,false,colors);
						}

						arm1+=pivotPoint;
						pivotPoint=spOutputVertices->spatialVector(
								index,subIndex-1);
						arm1-=pivotPoint;
					}
				}
			}
			spOutputVertices->set(index,subIndex,point);

			arm2=point-pivotPoint;
#if FE_DODGE_DEBUG
			feLog("pivotQuat %s\n",c_print(pivotQuat));
			feLog("pivotPoint %s\n",c_print(pivotPoint));
			feLog("arm1 %s\n",c_print(arm1));
			feLog("arm2 %s\n",c_print(arm2));
#endif
			if(changed)
			{
				set(pivotQuat,unitSafe(arm1),unitSafe(arm2));
				hit=TRUE;
			}
#if FE_DODGE_DEBUG
			feLog("pivotQuat %s\n",c_print(pivotQuat));
#endif
			lastPoint=oldPoint;
		}
	}
}
